<?php

namespace Drupal\vais\Commands;

use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\Utility\TokenizerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\vais\Utility\VaisLlmUtility;
use Drush\Commands\DrushCommands;

/**
 * Class VaisDrushCommands.
 *
 * Provides Drush commands for the VAIS module.
 */
class VaisDrushCommands extends DrushCommands {

  /**
   * Convert Markup into HTML.
   *
   * @var \Parsedown
   */
  protected \Parsedown $parseDown;

  /**
   * Constructs a new VaisDrushCommands object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\ai\AiProviderPluginManager $aiProvider
   *   The AI provider plugin manager.
   * @param \Drupal\ai\Utility\TokenizerInterface $tokenizer
   *   The tokenizer service.
   * @param \Drupal\vais\Utility\VaisLlmUtility $llmUtility
   *   The LLM utility service.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected AiProviderPluginManager $aiProvider,
    protected TokenizerInterface $tokenizer,
    protected VaisLlmUtility $llmUtility,
  ) {
    parent::__construct();
    $this->parseDown = new \Parsedown();
  }

  /**
   * Custom Drush command to update a field on entities.
   *
   * @param array $options
   *   An associative array of options.
   *
   * @option llm
   *   The LLM to use for summarizing.
   * @option context_window
   *   Maximum number of tokens which the model can handle.
   * @option summary_length
   *   Required length of the resulted summary in tokens.
   * @option entity_type
   *   The entity type (e.g., 'node').
   * @option bundle
   *   The bundle (e.g., 'article').
   * @option input_field_name
   *   The field name to update (e.g., 'field_custom').
   * @option output_field_name
   *   The value to assign to the field.
   * @option eids
   *   (Optional) A comma-separated list of entity IDs to update.
   * @option taxonomy_field_name
   *   (Optional) The field where taxonomy terms are stored.
   * @option vocabulary
   *   (Optional) The vocabulary machine name (e.g., 'tags').
   * @option term_name
   *   (Optional) The taxonomy term name (e.g., 'News').
   * @option llm_prompt
   *   (Optional) The prompt used in LLM call.
   * @option retain_existing
   *   (Optional) if True, the output_field will be overwritten only if empty.
   *
   * @command vais:ai-summary-field
   * @aliases vais-ai-summary-field
   * @usage vais:ai-summary-field --llm=openai__gpt-4o --context_window=60000
   *   --entity_type=node --bundle=wiki_record --eids="1608"
   *   --input_field_name=body --output_field_name=field_ai_summary
   *   --retain_existing Updates the field_ai_summary on node of type
   *   wiki_record, NID 1608, with the summary of the body field generated
   *   by OpenAI's GPT-4o, if field_ai_summary is not empty.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function updateField(array $options = [
    'llm' => NULL,
    'context_window' => NULL,
    'summary_length' => 150,
    'entity_type' => NULL,
    'bundle' => NULL,
    'eids' => NULL,
    'input_field_name' => NULL,
    'output_field_name' => NULL,
    'taxonomy_field_name' => NULL,
    'vocabulary' => NULL,
    'term_name' => NULL,
    'llm_prompt' => NULL,
    'retain_existing' => FALSE,
  ]): void {
    $llm = $options['llm'];
    $context_window = $options['context_window'];
    $summary_length = $options['summary_length'];
    $entity_type = $options['entity_type'];
    $bundle = $options['bundle'];
    $eids = $options['eids'];
    $input_field_name = $options['input_field_name'];
    $output_field_name = $options['output_field_name'];
    $taxonomy_field_name = $options['taxonomy_field_name'];
    $vocabulary = $options['vocabulary'];
    $term_name = $options['term_name'];
    $llm_prompt = $options['llm_prompt'];
    $retain_existing = $options['retain_existing'];

    // Ensure we have LLM provider available.
    if (!$this->aiProvider->hasProvidersForOperationType('chat')) {
      $this->logger()->warning(dt('No LLM is available, please @install and @configure one first.', [
        '@install' => Link::createFromRoute('install', 'system.modules_list')->toString(),
        '@configure' => Link::createFromRoute('configure', 'ai.admin_providers')->toString(),
      ]));
      return;
    }

    // Ensure we have LLM model selected.
    if (!$llm) {
      $this->logger()->warning(
        dt(
          "You need to provide an LLM to use.\n Choose one of these:\n@llms",
          [
            '@llms' => $this->showAvailableModels(),
          ]
        )
      );
      return;
    }

    // Ensure we have both fields.
    if (!($input_field_name && $output_field_name)) {
      $this->logger()->warning(dt('You need to provide input and output fields.'));
      return;
    }

    // Build the entity query.
    if ($entity_type && $bundle) {
      $query = $this->entityTypeManager->getStorage($entity_type)->getQuery()
        ->condition('type', $bundle)
        ->accessCheck(FALSE);
    }
    else {
      $this->logger()->warning(dt('You need to provide an entity and its bundle.'));
      return;
    }

    // If we have set Entity IDs, use these as filter.
    if ($eids) {
      $entity_type_definition = $this->entityTypeManager->getDefinition($entity_type);
      $id_field = $entity_type_definition->getKey('id');
      $eid_array = array_map('trim', explode(',', $eids));
      $query->condition($id_field, $eid_array, 'IN');
    }

    // If vocabulary and term name are provided, load the term.
    if ($vocabulary && $term_name && $taxonomy_field_name) {
      $term = $this->entityTypeManager->getStorage('taxonomy_term')->loadByProperties([
        'name' => $term_name,
        'vid' => $vocabulary,
      ]);
      $term = reset($term);
      if ($term) {
        $term_id = $term->id();
        $query->condition('field_tags', $term_id, 'IN');
      }
      else {
        $this->logger()->warning(dt('Taxonomy term "@term_name" in vocabulary "@vocabulary" not found.', [
          '@term_name' => $term_name,
          '@vocabulary' => $vocabulary,
        ]));
        return;
      }
    }

    // Execute query and obtain entities.
    $ids = $query->execute();
    if (empty($ids)) {
      $this->logger()->notice(dt('No entities found.'));
      return;
    }

    // Initialize tokenizer.
    try {
      $this->tokenizer->setModel(
        $this->aiProvider->getModelNameFromSimpleOption($llm)
      );
    }
    catch (\InvalidArgumentException) {
      $this->tokenizer->setModel('gpt-3.5-turbo');
    }

    // Update entities.
    foreach ($ids as $id) {
      $entity = $this->entityTypeManager->getStorage($entity_type)->load($id);
      if ($entity) {

        // Skip entity with existing content in output if required.
        if ($retain_existing && trim($entity->get($output_field_name)->value)) {
          $this->logger()->notice(
            dt('Skipped entity ID @eid. Not empty.', ['@eid' => $id])
          );
          continue;
        }

        // Summarize.
        $input_field = $this->llmUtility->sanitize($entity->get($input_field_name)->value);
        $tokens_sum = $this->tokenizer->countTokens($input_field);
        if ($tokens_sum <= $summary_length) {
          $value = $input_field;
          $this->logger()->notice(
            dt(
              'Field @field in entity ID @eid was copied as is.',
              [
                '@eid' => $id,
                '@field' => $input_field_name,
              ]
            )
          );
        }
        elseif ($tokens_sum + 250 < $context_window) {
          $value = $this->summarize($input_field, $llm_prompt, $llm);
          $this->logger()->notice(
            dt(
              'Field @field in entity ID @eid was summarized.',
              [
                '@eid' => $id,
                '@field' => $input_field_name,
              ]
            )
                  );
        }
        else {
          $this->logger()->notice(
            dt(
              "Content of @field in entity ID @eid doesn't fit context window.",
              [
                '@eid' => $id,
                '@field' => $input_field_name,
              ]
            )
                  );
          continue;
        }
        $entity->set(
          $output_field_name,
          [
            'value' => $this->parseDown->text($value),
            'format' => 'full_html',
          ]
        );
        $entity->save();
      }
    }
  }

  /**
   * Return formatted list of available LLMs.
   *
   * @return string
   *   List of available models.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function showAvailableModels(): string {
    $models = '';
    foreach ($this->aiProvider->getProvidersForOperationType('chat') as $id => $definition) {
      $provider = $this->aiProvider->createInstance($id);
      foreach ($provider->getConfiguredModels('chat') as $model => $label) {
        $models .= ' - For ' . $definition['label']->__toString() . ' | ' . $label . ' use "' . $id . '__' . $model . "\"\n";
      }
    }
    return $models;
  }

  /**
   * Summarize given text.
   *
   * @param string $text
   *   Text to summarize.
   * @param string|null $llm_prompt
   *   Prompt for LLM.
   * @param string $llm
   *   LLM provider as set in the AI module.
   *
   * @return mixed
   *   Summarized text.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function summarize(string $text, string|null $llm_prompt, string $llm): mixed {
    if (!$llm_prompt) {
      $prompt = <<<EOD
Summarize the following text in a concise manner and in language in which it is written.
The summary should:
	-	Be no longer than 150 tokens.
	-	Highlight the most critical points and key takeaways.

Use only the provided text, DO NOT rely upon previous knowledge.

Text to summary:

EOD;
    }
    else {
      $prompt = $llm_prompt;
    }
    $prompt .= $text;
    return $this->llmUtility->getResponse($prompt, $llm);
  }

}
