<?php

namespace Drupal\vais\Plugin\Block;

use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\Utility\TokenizerInterface;
use Drupal\Core\Block\Annotation\Block;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\vais\Utility\VaisLlmUtility;
use Drupal\views\Plugin\views\filter\FilterPluginBase;
use Drupal\views\ViewExecutable;
use Drupal\views\Views;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides a 'Views AI Summary' Block.
 *
 * @Block(
 *   id = "vais_block",
 *   admin_label = @Translation("Views AI Summary"),
 * )
 */
class VaisViewBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * Constructs a new VaisViewBlock instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\ai\AiProviderPluginManager $aiProviderManager
   *   The AI provider manager service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\vais\Utility\VaisLlmUtility $llmUtility
   *   The LLM utility service.
   * @param \Drupal\ai\Utility\TokenizerInterface $tokenizer
   *   The tokenizer service.
   * @param \Drupal\Core\Session\AccountProxyInterface $currentUser
   *   The current user service.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected AiProviderPluginManager $aiProviderManager,
    protected RequestStack $requestStack,
    protected RouteMatchInterface $route_match,
    protected LoggerInterface $logger,
    protected VaisLlmUtility $llmUtility,
    protected TokenizerInterface $tokenizer,
    protected AccountProxyInterface $currentUser
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('ai.provider'),
      $container->get('request_stack'),
      $container->get('current_route_match'),
      $container->get('logger.factory')->get('vais'),
      $container->get('vais.llm_utility'),
      $container->get('ai.tokenizer'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function build(): array {

    // Ensure only allowed users can access llm.
    if (!$this->currentUser->hasPermission('access summary vais')) {
      $this->logger->warning(
        'User @uid (@username) does not have permission to access the summary block.',
        [
          '@uid' => $this->currentUser->id(),
          '@username' => $this->currentUser->getAccountName(),
        ]
      );
      return [];
    }

    $route_name = $this->route_match->getRouteName();
    if (empty($this->configuration['llm_engine'])) {
      return [
        '#markup' => $this->t("No LLM engine for summary has been selected yet."),
      ];
    }
    else {
      $output = $this->t("This page is not a View.");
    }

    if (str_starts_with($route_name, 'view.')) {
      $view_id = $this->route_match->getParameter('view_id');
      $display_id = $this->route_match->getParameter('display_id');
      if ($view_id && $display_id) {
        $view = Views::getView($view_id);
        if ($view) {
          $view->setDisplay($display_id);
          $view->execute();
          $num_rows = $this->configuration['num_rows'] ?? 5;
          $process_view_fields = $this->configuration['process_view_fields'];
          $rendered_rows = [];
          $composite_content = '';
          foreach ($view->result as $row_index => $row) {
            if ($row_index == $num_rows) {
              break;
            }
            $rendered_fields = [];
            foreach ($view->field as $field) {
              $excluded = !empty($field->options['exclude']);
              $add_excluded = in_array(
                  $process_view_fields,
                  ['only_excluded', 'both']
                ) && $excluded;
              $add_rendered = in_array(
                  $process_view_fields,
                  ['only_rendered', 'both']
                ) && !$excluded;
              if ($add_excluded || $add_rendered) {
                $content_utf8 = $this->llmUtility->sanitize((string) $field->render($row));
                $rendered_fields[] = [
                  'content' => $content_utf8,
                  'label' => $field->options['label'],
                ];
              }
            }
            $rendered_rows[] = $rendered_fields;
          }

          // Initialize tokenizer.
          [, $llm_model_id] = explode('__', $this->configuration['llm_engine']);
          try {
            $this->tokenizer->setModel($llm_model_id);
          }
          catch (\InvalidArgumentException) {
            $this->tokenizer->setModel('gpt-3.5-turbo');
          }

          // Process View rows.
          foreach ($rendered_rows as $idx => $rendered_row) {
            $composite_content .= ($idx + 1) . ' SEARCH RESULT ROW.';
            $composite_content .= "\n----------\n\n";
            foreach ($rendered_row as $rendered_field) {
              if (!empty($rendered_field['label'])) {
                $composite_content .= $rendered_field['label'] . ': ';
              }
              $composite_content .= $rendered_field['content'] . "\n\n";
              if ($this->tokenizer->countTokens($composite_content) >= ($this->configuration['context_length'] * .8)) {
                break 2;
              }
            }
          }

          $user_prompt = $this->getSearchQueryString($view);
          $llm_prompt = <<<EOD
Below are a results of a search, ordered by importance."
EOD;
          if ($user_prompt) {
            $llm_prompt .= "\n\nThe search query was: \"$user_prompt\"\n\n";
            $llm_prompt .= <<<EOD
Summarize the search results in few sentences, up-to two paragraphs, but produce no more than 150 tokens.
Feel free to use bullet points and/or ordered lists.
Use only the search results, DO NOT rely upon previous knowledge.

RESULTS:
EOD;
          }
          else {
            $llm_prompt .= <<<EOD
Summarize the search result in few sentences, up-to two paragraphs, but no more than 150 tokens.
Feel free to use bullet points or ordered lists.
Do not use direct reference to SEARCH RESULT ROW, but you can quote.

Use only the search results, DO NOT rely upon previous knowledge.

RESULTS:
EOD;
          }

          if (trim(strip_tags($composite_content))) {
            $placeholder_text = $this->configuration['placeholder_text'] ?? 'Creating summary';
            $llm_prompt .= "\n\n$composite_content";
            $build = [
              '#markup' => '<div id="summary-placeholder">' . $placeholder_text . '<span class="dots"></span></div>',
              '#attached' => [
                'library' => [
                  'vais/summary_loader',
                ],
                'drupalSettings' => [
                  'vaisPrompt' => $llm_prompt,
                  'llm' => $this->configuration['llm_engine'],
                ],
              ],
            ];
            if (!empty($this->configuration['add_animation'])) {
              $build['#attached']['library'][] = 'vais/summary_animation';
            }
            return $build;
          }
          else {
            return [];
          }
        }
      }
    }

    return [
      '#markup' => $output,
    ];
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function blockForm($form, FormStateInterface $form_state): array {
    $form = parent::blockForm($form, $form_state);

    // If no provider is installed we can't do anything.
    if (!$this->aiProviderManager->hasProvidersForOperationType('chat')) {
      $form['markup'] = [
        '#markup' => '<div class="ai-error">' . $this->t('No LLM is available, please %install and %configure one first.', [
          '%install' => Link::createFromRoute($this->t('install'), 'system.modules_list')->toString(),
          '%configure' => Link::createFromRoute($this->t('configure'), 'ai.admin_providers')->toString(),
        ]) . '</div>',
      ];
      return $form;
    }

    $form['num_rows'] = [
      '#type' => 'select',
      '#title' => $this->t('Number of rows to summarize'),
      '#description' => $this->t('Select the number of rows which will be summarized by AI (1-5).'),
      '#options' => array_combine(range(1, 5), range(1, 5)),
      '#default_value' => $this->configuration['num_rows'] ?? 5,
    ];

    $form['process_view_fields'] = [
      '#type' => 'radios',
      '#title' => $this->t('Process View Fields'),
      '#description' => $this->t('Select which fields to process.'),
      '#options' => [
        'only_rendered' => $this->t('Only rendered'),
        'only_excluded' => $this->t('Only excluded'),
        'both' => $this->t('Both'),
      ],
      '#default_value' => $this->configuration['process_view_fields'] ?? 'both',
    ];

    $form['llm_engine'] = [
      '#type' => 'select',
      '#title' => $this->t('LLM engine for summary'),
      '#options' => $this->getLlmOptions(),
      '#required' => TRUE,
      '#description' => $this->t('The service to use for creation of summary.'),
    ];
    if (isset($this->configuration['llm_engine'])) {
      $form['llm_engine']['#default_value'] = $this->configuration['llm_engine'];
    }

    // @todo Define #max if provided by the LLM. This is NOT max_tokens!
    $form['context_length'] = [
      '#type' => 'number',
      '#title' => $this->t('Context Length'),
      '#description' => $this->t('Enter the maximum length of the LLMs (input) context window you want to use.'),
      '#default_value' => $this->configuration['context_length'] ?? 1024,
      '#min' => 1,
    ];

    $form['add_animation'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Add animation to placeholder'),
      '#description' => $this->t('Check this box to add an animated ellipsis to the placeholder.'),
      '#default_value' => $this->configuration['add_animation'] ?? FALSE,
    ];

    $form['placeholder_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Placeholder Text'),
      '#description' => $this->t('Enter the text to display as the placeholder while the summary is being created.'),
      '#default_value' => $this->configuration['placeholder_text'] ?? 'Creating summary',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state): void {
    parent::blockSubmit($form, $form_state);
    $this->configuration['num_rows'] = $form_state->getValue('num_rows');
    $this->configuration['process_view_fields'] = $form_state->getValue('process_view_fields');
    $this->configuration['llm_engine'] = $form_state->getValue('llm_engine');
    $this->configuration['context_length'] = $form_state->getValue('context_length');
    $this->configuration['add_animation'] = $form_state->getValue('add_animation');
    $this->configuration['placeholder_text'] = $form_state->getValue('placeholder_text');
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge(): int {
    return 0;
  }

  /**
   * Returns all available LLM chat engines as options.
   *
   * @return array
   *   The LLM engines.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function getLlmOptions(): array {
    $options = [];
    foreach ($this->aiProviderManager->getProvidersForOperationType('chat') as $id => $definition) {
      $provider = $this->aiProviderManager->createInstance($id);
      foreach ($provider->getConfiguredModels('chat') as $model => $label) {
        $options[$id . '__' . $model] = $definition['label']->__toString() . ' | ' . $label;
      }
    }
    if (empty($options)) {
      $this->messenger->addWarning('No LLM available. Please install and enable an LLM module provider before continuing.');
    }
    return $options;
  }

  /**
   * Returns search query string if it is set.
   *
   * @param \Drupal\views\ViewExecutable $view
   *   A View.
   *
   * @return string
   *   User query string.
   */
  private function getSearchQueryString(ViewExecutable $view): string {
    $request = $this->requestStack->getCurrentRequest();
    $query_params = $request->query->all();

    foreach ($view->display_handler->getHandlers('filter') as $filter) {
      if ($filter instanceof FilterPluginBase && $filter->isExposed() && $filter->getPluginId() === 'search_api_fulltext') {
        $identifier = $filter->options['expose']['identifier'];
        if (isset($query_params[$identifier])) {
          return (string) $query_params[$identifier];
        }
      }
    }

    return '';
  }

}
