<?php

namespace Drupal\vais\Utility;

use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\OperationType\Chat\ChatInput;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\ai\Plugin\ProviderProxy;
use Drupal\Component\Utility\Html;
use League\HTMLToMarkdown\Converter\TableConverter;
use League\HTMLToMarkdown\HtmlConverter;

/**
 * Utility class for interacting with LLM providers.
 */
class VaisLlmUtility {

  /**
   * HTML to markdown converter.
   *
   * @var \League\HTMLToMarkdown\HtmlConverter
   */
  protected HtmlConverter $converter;

  /**
   * The LLM provider instance.
   *
   * @var \Drupal\ai\Plugin\ProviderProxy
   */
  protected ProviderProxy $llm;

  /**
   * Constructs a new VaisLlmUtility object.
   *
   * @param \Drupal\ai\AiProviderPluginManager $aiProvider
   *   The AI provider plugin manager.
   */
  public function __construct(
    protected AiProviderPluginManager $aiProvider
  ) {
    $this->converter = new HtmlConverter();
    $this->converter->getConfig()->setOption('strip_tags', TRUE);
    $this->converter->getConfig()->setOption('strip_placeholder_links', TRUE);
    $this->converter->getEnvironment()->addConverter(new TableConverter());
  }

  /**
   * Returns LLM Provider.
   *
   * @param string $llm_engine
   *   The LLM engine configuration.
   *
   * @return \Drupal\ai\Plugin\ProviderProxy
   *   Returns LLM provider.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function getLlmProvider(string $llm_engine): ProviderProxy {
    [$llm_provider_id] = explode('__', $llm_engine);

    if (empty($this->llm)) {
      $this->llm = $this->aiProvider->createInstance($llm_provider_id);
    }

    return $this->llm;
  }

  /**
   * Returns response from LLM for given input.
   *
   * @param string $input
   *   The input for the LLM.
   * @param string $llm_engine
   *   The LLM engine configuration.
   *
   * @return mixed
   *   The normalized response from the LLM.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function getResponse(string $input, string $llm_engine): mixed {
    [, $llm_model_id] = explode('__', $llm_engine);
    $config = [
      "temperature" => 0,
    ];
    $input = new ChatInput(
      [
        new ChatMessage("system", "You are an expert summarizer."),
        new ChatMessage("user", $input),
      ]
    );
    $llm = $this->getLlmProvider($llm_engine);
    $llm->setConfiguration($config);
    return $llm
      ->chat($input, $llm_model_id, ["vais"])
      ->getNormalized()
      ->getText();
  }

  /**
   * Sanitize input text.
   *
   * @param string $text
   *   Text to sanitize.
   *
   * @return string
   *   Sanitized text.
   */
  public function sanitize(string $text): string {
    $content_string = Html::decodeEntities($text);
    $content_utf8 = mb_convert_encoding($content_string, 'UTF-8', 'UTF-8');
    return $this->converter->convert($content_utf8);
  }

}
